--- servers/slapd/overlays/syncprov.c 2007/11/02 15:51:13 1.209
+++ servers/slapd/overlays/syncprov.c 2008/03/19 23:31:42 1.223
@@ -1,8 +1,8 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/syncprov.c,v 1.208 2007/11/02 15:40:42 hyc Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/syncprov.c,v 1.222 2008/03/19 23:26:14 hyc Exp $ */
/* syncprov.c - syncrepl provider */
/* This work is part of OpenLDAP Software .
*
- * Copyright 2004-2007 The OpenLDAP Foundation.
+ * Copyright 2004-2008 The OpenLDAP Foundation.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
@@ -696,7 +696,10 @@ again:
break;
}
- fop.o_bd->bd_info = on->on_info->oi_orig;
+ if ( on->on_next )
+ fop.o_bd->bd_info = (BackendInfo *)on->on_next;
+ else
+ fop.o_bd->bd_info = on->on_info->oi_orig;
fop.o_bd->be_search( &fop, &frs );
fop.o_bd->bd_info = (BackendInfo *)on;
@@ -837,8 +840,10 @@ syncprov_sendresp( Operation *op, opcook
/* Play back queued responses */
static int
-syncprov_qplay( Operation *op, slap_overinst *on, syncops *so )
+syncprov_qplay( Operation *op, struct re_s *rtask )
{
+ syncops *so = rtask->arg;
+ slap_overinst *on = LDAP_SLIST_FIRST(&so->s_op->o_extra)->oe_key;
syncres *sr;
Entry *e;
opcookie opc;
@@ -853,10 +858,10 @@ syncprov_qplay( Operation *op, slap_over
so->s_res = sr->s_next;
if ( !so->s_res )
so->s_restail = NULL;
- ldap_pvt_thread_mutex_unlock( &so->s_mutex );
-
+ /* Exit loop with mutex held */
if ( !sr || so->s_op->o_abandon )
break;
+ ldap_pvt_thread_mutex_unlock( &so->s_mutex );
opc.sdn = sr->s_dn;
opc.sndn = sr->s_ndn;
@@ -883,9 +888,24 @@ syncprov_qplay( Operation *op, slap_over
ch_free( sr );
- if ( rc )
+ if ( rc ) {
+ /* Exit loop with mutex held */
+ ldap_pvt_thread_mutex_lock( &so->s_mutex );
break;
+ }
}
+
+ /* wait until we get explicitly scheduled again */
+ ldap_pvt_thread_mutex_lock( &slapd_rq.rq_mutex );
+ ldap_pvt_runqueue_stoptask( &slapd_rq, rtask );
+ if ( rc == 0 ) {
+ ldap_pvt_runqueue_resched( &slapd_rq, rtask, 1 );
+ } else {
+ /* bail out on any error */
+ ldap_pvt_runqueue_remove( &slapd_rq, rtask );
+ }
+ ldap_pvt_thread_mutex_unlock( &slapd_rq.rq_mutex );
+ ldap_pvt_thread_mutex_unlock( &so->s_mutex );
return rc;
}
@@ -895,7 +915,6 @@ syncprov_qtask( void *ctx, void *arg )
{
struct re_s *rtask = arg;
syncops *so = rtask->arg;
- slap_overinst *on = so->s_op->o_private;
OperationBuffer opbuf;
Operation *op;
BackendDB be;
@@ -917,25 +936,14 @@ syncprov_qtask( void *ctx, void *arg )
be = *so->s_op->o_bd;
be.be_flags |= SLAP_DBFLAG_OVERLAY;
op->o_bd = &be;
- op->o_private = NULL;
+ LDAP_SLIST_FIRST(&op->o_extra) = NULL;
op->o_callback = NULL;
- rc = syncprov_qplay( op, on, so );
+ rc = syncprov_qplay( op, rtask );
/* decrement use count... */
syncprov_free_syncop( so );
- /* wait until we get explicitly scheduled again */
- ldap_pvt_thread_mutex_lock( &slapd_rq.rq_mutex );
- ldap_pvt_runqueue_stoptask( &slapd_rq, rtask );
- if ( rc == 0 ) {
- ldap_pvt_runqueue_resched( &slapd_rq, rtask, 1 );
- } else {
- /* bail out on any error */
- ldap_pvt_runqueue_remove( &slapd_rq, rtask );
- }
- ldap_pvt_thread_mutex_unlock( &slapd_rq.rq_mutex );
-
#if 0 /* FIXME: connection_close isn't exported from slapd.
* should it be?
*/
@@ -1209,7 +1217,7 @@ syncprov_matchops( Operation *op, opcook
}
/* check if current o_req_dn is in scope and matches filter */
- if ( fc.fscope && test_filter( op, e, ss->s_op->ors_filter ) ==
+ if ( fc.fscope && test_filter( ss->s_op, e, ss->s_op->ors_filter ) ==
LDAP_COMPARE_TRUE ) {
if ( saveit ) {
sm = op->o_tmpalloc( sizeof(syncmatches), op->o_tmpmemctx );
@@ -1301,6 +1309,7 @@ syncprov_checkpoint( Operation *op, Slap
Operation opm;
SlapReply rsm = { 0 };
slap_callback cb = {0};
+ BackendDB be;
mod.sml_numvals = si->si_numcsns;
mod.sml_values = si->si_ctxcsn;
@@ -1316,8 +1325,12 @@ syncprov_checkpoint( Operation *op, Slap
opm.o_callback = &cb;
opm.orm_modlist = &mod;
opm.orm_no_opattrs = 1;
- opm.o_req_dn = op->o_bd->be_suffix[0];
- opm.o_req_ndn = op->o_bd->be_nsuffix[0];
+ if ( SLAP_GLUE_SUBORDINATE( op->o_bd )) {
+ be = *on->on_info->oi_origdb;
+ opm.o_bd = &be;
+ }
+ opm.o_req_dn = opm.o_bd->be_suffix[0];
+ opm.o_req_ndn = opm.o_bd->be_nsuffix[0];
opm.o_bd->bd_info = on->on_info->oi_orig;
opm.o_managedsait = SLAP_CONTROL_NONCRITICAL;
opm.o_no_schema_check = 1;
@@ -1325,7 +1338,6 @@ syncprov_checkpoint( Operation *op, Slap
if ( mod.sml_next != NULL ) {
slap_mods_free( mod.sml_next, 1 );
}
- opm.orm_no_opattrs = 0;
}
static void
@@ -1519,7 +1531,10 @@ syncprov_playlog( Operation *op, SlapRep
fop.ors_filter = ⁡
cb.sc_response = playlog_cb;
- fop.o_bd->bd_info = on->on_info->oi_orig;
+ if ( on->on_next )
+ fop.o_bd->bd_info = (BackendInfo *)on->on_next;
+ else
+ fop.o_bd->bd_info = on->on_info->oi_orig;
for ( i=ndel; isi_chktime &&
(op->o_time - si->si_chklast >= si->si_chktime )) {
- do_check = 1;
- si->si_chklast = op->o_time;
+ if ( si->si_chklast ) {
+ do_check = 1;
+ si->si_chklast = op->o_time;
+ } else {
+ si->si_chklast = 1;
+ }
}
}
ldap_pvt_thread_rdwr_wunlock( &si->si_csn_rwlock );
@@ -1771,7 +1790,13 @@ syncprov_op_mod( Operation *op, SlapRepl
/* wait for this op to get to head of list */
while ( mt->mt_mods != mi ) {
ldap_pvt_thread_mutex_unlock( &mt->mt_mutex );
- ldap_pvt_thread_yield();
+ /* FIXME: if dynamic config can delete overlays or
+ * databases we'll have to check for cleanup here.
+ * Currently it's not an issue because there are
+ * no dynamic config deletes...
+ */
+ if ( !ldap_pvt_thread_pool_pausecheck( &connection_pool ))
+ ldap_pvt_thread_yield();
ldap_pvt_thread_mutex_lock( &mt->mt_mutex );
/* clean up if the caller is giving up */
@@ -1839,6 +1864,7 @@ syncprov_search_cleanup( Operation *op,
typedef struct SyncOperationBuffer {
Operation sob_op;
Opheader sob_hdr;
+ OpExtra sob_oe;
AttributeName sob_extra; /* not always present */
/* Further data allocated here */
} SyncOperationBuffer;
@@ -1867,6 +1893,7 @@ syncprov_detach_op( Operation *op, synco
sopbuf2 = ch_calloc( 1, size );
op2 = &sopbuf2->sob_op;
op2->o_hdr = &sopbuf2->sob_hdr;
+ LDAP_SLIST_FIRST(&op2->o_extra) = &sopbuf2->sob_oe;
/* Copy the fields we care about explicitly, leave the rest alone */
*op2->o_hdr = *op->o_hdr;
@@ -1874,7 +1901,8 @@ syncprov_detach_op( Operation *op, synco
op2->o_time = op->o_time;
op2->o_bd = on->on_info->oi_origdb;
op2->o_request = op->o_request;
- op2->o_private = on;
+ LDAP_SLIST_FIRST(&op2->o_extra)->oe_key = on;
+ LDAP_SLIST_NEXT(LDAP_SLIST_FIRST(&op2->o_extra), oe_next) = NULL;
ptr = (char *) sopbuf2 + offsetof( SyncOperationBuffer, sob_extra );
if ( i ) {
@@ -1925,12 +1953,10 @@ syncprov_detach_op( Operation *op, synco
op2->o_do_not_cache = 1;
/* Add op2 to conn so abandon will find us */
- ldap_pvt_thread_mutex_lock( &op->o_conn->c_mutex );
op->o_conn->c_n_ops_executing++;
op->o_conn->c_n_ops_completed--;
LDAP_STAILQ_INSERT_TAIL( &op->o_conn->c_ops, op2, o_next );
so->s_flags |= PS_IS_DETACHED;
- ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
/* Prevent anyone else from trying to send a result for this op */
op->o_abandon = 1;
@@ -2040,15 +2066,27 @@ syncprov_search_response( Operation *op,
/* Detach this Op from frontend control */
ldap_pvt_thread_mutex_lock( &ss->ss_so->s_mutex );
+ ldap_pvt_thread_mutex_lock( &op->o_conn->c_mutex );
+
+ /* But not if this connection was closed along the way */
+ if ( op->o_abandon ) {
+ ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
+ ldap_pvt_thread_mutex_unlock( &ss->ss_so->s_mutex );
+ syncprov_free_syncop( ss->ss_so );
+ return SLAPD_ABANDON;
+
+ } else {
+ /* Turn off the refreshing flag */
+ ss->ss_so->s_flags ^= PS_IS_REFRESHING;
- /* Turn off the refreshing flag */
- ss->ss_so->s_flags ^= PS_IS_REFRESHING;
+ syncprov_detach_op( op, ss->ss_so, on );
- syncprov_detach_op( op, ss->ss_so, on );
+ ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
- /* If there are queued responses, fire them off */
- if ( ss->ss_so->s_res )
- syncprov_qstart( ss->ss_so );
+ /* If there are queued responses, fire them off */
+ if ( ss->ss_so->s_res )
+ syncprov_qstart( ss->ss_so );
+ }
ldap_pvt_thread_mutex_unlock( &ss->ss_so->s_mutex );
return LDAP_SUCCESS;
@@ -2321,6 +2359,12 @@ syncprov_operational(
slap_overinst *on = (slap_overinst *)op->o_bd->bd_info;
syncprov_info_t *si = (syncprov_info_t *)on->on_bi.bi_private;
+ /* This prevents generating unnecessarily; frontend will strip
+ * any statically stored copy.
+ */
+ if ( op->o_sync != SLAP_CONTROL_NONE )
+ return SLAP_CB_CONTINUE;
+
if ( rs->sr_entry &&
dn_match( &rs->sr_entry->e_nname, op->o_bd->be_nsuffix )) {
@@ -2840,10 +2884,14 @@ static int syncprov_parseCtrl (
sr->sr_rhint = rhint;
if (!BER_BVISNULL(&cookie)) {
ber_dupbv_x( &sr->sr_state.octet_str, &cookie, op->o_tmpmemctx );
+ /* If parse fails, pretend no cookie was sent */
if ( slap_parse_sync_cookie( &sr->sr_state, op->o_tmpmemctx ) ||
sr->sr_state.rid == -1 ) {
- rs->sr_text = "Sync control : cookie parsing error";
- return LDAP_PROTOCOL_ERROR;
+ if ( sr->sr_state.ctxcsn ) {
+ ber_bvarray_free_x( sr->sr_state.ctxcsn, op->o_tmpmemctx );
+ sr->sr_state.ctxcsn = NULL;
+ }
+ sr->sr_state.numcsns = 0;
}
}